@@ -1,126 +1,52 @@
-########################################################################
-# FindBin::libs
-#
-# use $FindBin::Bin to search for 'lib' directories and use them.
-#
-# default action is to look for dir's named "lib" and silently use
-# the lib's without exporting anything. print turns on a short 
-# message with the abs_path results, export pushes out a variable
-# (default name is the base value), verbose turns on decision output
-# and print. export takes an optional argument with the name of a
-# variable to export.
-#
-# Copyright (C) 2003-2012, Steven Lembark, Workhorse Computing.
-# This code is released under the same terms as Perl-5.8
-# or any later version of Perl.
-# 
-########################################################################
-
-########################################################################
+#######################################################################
 # housekeeping
 ########################################################################
 
-package FindBin::libs;
-
-use v5.8;
-use strict;
+package FindBin::libs   v3.0.0;
+use v5.28;
 
 use FindBin;
-use Symbol;
-
 use File::Basename;
 
-use Carp    qw( croak           );
-use Symbol  qw( qualify_to_ref  );
+use Carp    qw( croak                   );
+use Cwd     qw( abs_path                );
+use Symbol  qw( qualify qualify_to_ref  );
 
 use File::Spec::Functions
 qw
 (
-    &splitpath
-    &splitdir
-    &catpath
-    &catdir
+    splitpath
+    splitdir
+    catpath
+    catdir
+    canonpath
 );
 
-BEGIN
-{
-    # however... there have been complaints of 
-    # places where abs_path does not work. 
-    #
-    # if abs_path fails on the working directory
-    # then replace it with rel2abs and live with 
-    # possibly slower, redundant directories.
-    #
-    # the abs_path '//' hack allows for testing 
-    # broken abs_path on primitive systems that
-    # cannot handle the rooted system being linked
-    # back to itself.
-
-    use Cwd qw( &cwd );
-
-    my $abs = Cwd->can( 'abs_path'  )
-    or die "Odd: Cwd cannot 'abs_path'\n";
-
-    if
-    (
-        eval { $abs->( '//' );  $abs->( cwd ); 1 }
-    )
-    {
-        # nothing more to do: abs_path works.
-    }
-    elsif
-    (
-        $abs = File::Spec::Functions->can( 'rel2abs'  )
-    )
-    {
-        # ok, we have a substitute
-    }
-    else
-    {
-        die "Cwd fails abs_path test && lacks 'rel2abs'\n";
-    }
-
-    my $ref = qualify_to_ref 'abs_path', __PACKAGE__;
-
-    undef &{ *$ref };
-
-    *{ $ref } = $abs;
-}
-
 ########################################################################
 # package variables 
 ########################################################################
 
-our $VERSION = '1.062_01';
-$VERSION = eval $VERSION;
-
 my %defaultz = 
 (
-    base    => 'lib',
-    use     => undef,
+    base    => 'lib'
+  , use     => undef
+  , blib    => undef    # prefer ./blib at the first level
 
-    subdir  => '',      # add this subdir also if found.
-    subonly => undef,   # leave out lib's, use only subdir.
-    export  => undef,   # push variable into caller's space.
-    verbose => undef,   # boolean: print inputs, results.
-    debug   => undef,   # boolean: set internal breakpoints.
+  , subdir  => '',      # add this subdir also if found.
+  , subonly => undef    # leave out lib's, use only subdir.
+  , export  => undef    # push variable into caller's space.
+  , append  => undef    # push onto existing array (vs. overwrite)
+  , verbose => undef    # boolean: print inputs, results.
+  , debug   => undef    # boolean: set internal breakpoints.
 
-    print   => undef,   # display the results
+  , print   => 1        # display the results
 
-    p5lib   => undef,   # prefix PERL5LIB with the results
+  , p5lib   => undef    # prefix PERL5LIB with the results
 
-    ignore => '/,/usr', # dir's to skip looking for ./lib
+  # dir's to skip looking for ./lib
+  , ignore => [ qw( / /usr ) ]
 );
 
-# only new directories are used, ignore pre-loads
-# this with unwanted values.
-
-my %found = ();
-
-# saves passing this between import and $handle_args.
-
-my %argz    = ();
-my $verbose = '';
 my $empty   = q{};
 
 ########################################################################
@@ -132,14 +58,48 @@ my $empty   = q{};
 # directory path for it on VMS. Fix is to unshift a leading
 # '' into @dirpath where the leading entry is true.
 
-my $find_libs
+my $find_dirs
 = sub
 {
-    my $base    = basename ( shift || $argz{ base } );
+    my $argz    = shift;
+    my $base    = basename $argz->{ base };
+    my $subdir  
+    = $argz->{ subdir }
+    ? canonpath $argz->{ subdir }
+    : ''
+    ;
+    my $subonly = $argz->{ subonly } // '';
+    my %seen    = ();
+    my @libz    = ();
 
-    my $subdir  = $argz{ subdir } || '';
+    my $onepass = $argz->{ export } && $argz->{ scalar };
+    my $ignore  
+    = do
+    {
+        # ignoring nothing may be used with export,
+        # mainly for testing.
 
-    my $subonly = defined $argz{ subonly };
+        my $dirz    = $argz->{ ignore };
+        
+        if( @$dirz )
+        {
+            my $altz    
+            = join '|' =>
+            map
+            {
+                catdir $_, $base
+            }
+            @$dirz;
+
+            qr{^ (?: $altz ) $}x
+        }
+        else
+        {
+            qr{^$}
+        }
+    };
+
+    my $verbose = $argz->{ verbose };
 
     # for some reason, RH Enterprise V/4 has a 
     # trailing '/'; I havn't seen another copy of 
@@ -153,10 +113,10 @@ my $find_libs
     # after that splitpath can grab the directory 
     # portion for future use.
 
-    my ( $Bin ) = $argz{ Bin } =~ m{^ (.+) }xs;
+    my ( $Bin ) = ( $argz->{ Bin } =~ m{^ (.+) /? $}xs );
 
-    print STDERR "\nSearching $Bin for '$base'...\n"
-        if $verbose;
+    print STDERR "\nSearching $Bin for '$base' ($subdir)...\n"
+    if $verbose;
 
     my( $vol, $dir ) = splitpath $Bin, 1;
 
@@ -168,53 +128,74 @@ my $find_libs
 
     unshift @dirpath, '' if $dirpath[ 0 ];
 
-    my @libz    = ();
-
+    PATH:
     for( 1 .. @dirpath )
     {
         # note that catpath is extraneous on *NIX; the 
         # volume only means something on DOS- & VMS-based
         # filesystems, and adding an empty basename on 
         # *nix is unnecessary.
-        #
-        # HAK ALERT: the poor slobs stuck on windog have an
-        # abs_path that croaks on missing directories. have
-        # to eval the check for subdir's. 
 
-        my $abs
-        = eval { abs_path catpath $vol, ( catdir @dirpath, $base ), $empty }
-        || '';
+        my $dir = catpath $vol, ( catdir @dirpath, $base ), $empty;
 
-        my $sub
-        = $subdir
-        ? eval { abs_path ( catpath '', $abs, $subdir ) } || ''
-        : ''
-        ;
+        $dir    =~ $ignore
+        and next;
 
-        my @search = $subonly ? ( $sub ) : ( $abs, $sub );
+        -d $dir or next;
 
-        for my $dir ( @search )
+        # regardless of how things are symlinked, the abspath
+        # check only puts each directory on the list once.
+
+        my @dirz
+        = do
         {
-            if( $dir && -d $dir && ! exists $found{ $dir } )
+            if( $subdir )
             {
-                $found{ $dir } = 1;
+                my $sub = catpath $vol, ( catdir $dir, $subdir ), $empty;
 
-                push @libz, $dir;
+                if( -d $sub )
+                {
+                    $subonly
+                    ? ( $sub )
+                    : ( $sub, $dir )
+                }
+                else
+                {
+                    ()
+                }
+            }
+            else
+            {
+                ( $dir )
             }
         }
+        or next;
 
-        pop @dirpath
+        for my $dir ( @dirz )
+        {
+            $seen{ abs_path $dir } 
+            ||= push @libz, $dir;
+        }
+
+        last if $onepass;
+    }
+    continue
+    {
+        pop @dirpath;
     }
 
     # caller gets back the existing lib paths 
     # (including volume) walking up the path 
     # from $FindBin::Bin -> root.
     #
+    # no libs found is empty list or undef for 
+    # scalar.
+    #
     # passing it back as a list isn't all that
     # painful for a few paths.
 
     wantarray ? @libz : \@libz
-};;
+};
 
 # break out the messy part into a separate block.
 
@@ -228,20 +209,34 @@ my $handle_args
     # anything after the module are options with arguments
     # assigned via '='.
 
-    %argz
+    my %argz
     = map
     {
+        my $use_undef
+        = do
+        {
+            my %a   = ();
+            @a{ qw( export ignore ) } = ();
+            \%a
+        };
+
         my ( $k, $v ) = split '=', $_, 2;
 
-        $k =~ s{^ (?:!|no) }{}x
-        ? ( $k => undef )
-        : ( $k => ( $v || '' )  )
+        exists $use_undef->{ $k }
+        or $v //= 1;
+
+        # "no" inverts the sense of the test.
+
+        $k =~ s{^no}{}
+        and $v  = ! $v;
+
+        ( $k => $v )
     }
     @_;
 
     # stuff "debug=1" into your arguments and perl -d will stop here.
 
-    $DB::single = 1 if defined $argz{debug};
+    $DB::single = 1 if defined $argz{ debug };
 
     # default if nothing is supplied is to use the result;
     # otherwise, without use supplied either of export or
@@ -266,133 +261,163 @@ my $handle_args
     : $FindBin::Bin
     ;
 
+    if( exists $argz{ ignore } )
+    {
+        # ignore is different: it requires a bit of 
+        # compilation.
+        #
+        # ignore may be empty, mainly for testing.
+
+        if( my $dirz = $argz{ ignore } )
+        {
+            $argz{ ignore } = [ split ',' => $dirz ];
+        }
+        else
+        {
+            $argz{ ignore } = [];
+        }
+    }
+
     # now apply the defaults, then sanity check the result.
     # base is a special case since it always has to exist.
     #
-    # if $argz{export} is defined but false then it takes
-    # its default from $argz{base}.
+    # if $argz{ export } is defined but false then it takes
+    # its default from $argz{ base }.
 
-    exists $argz{$_} or $argz{$_} = $defaultz{$_}
-    for keys %defaultz;
+    while( my($k,$v) = each %defaultz )
+    {
+        # //= doesn't work here since undef may be a 
+        # legit default.
 
-    exists $argz{base} && $argz{base} 
+        exists $argz{ $k }
+        or
+        $argz{ $k } = $v;
+    }
+
+    exists $argz{ base } && $argz{ base } 
     or croak "Bogus FindBin::libs: missing/false base argument, should be 'base=NAME'";
 
-    defined $argz{export} and $argz{export} ||= $argz{base};
+    exists $argz{ export }
+    and
+    $argz{ export } //= $argz{ base };
 
-    $argz{ ignore } =
-    [
-        grep { $_ }
-        split /\s*,\s*/,
-        $argz{ignore}
-    ];
+    $argz{ verbose } //= '';
 
-    $verbose = defined $argz{verbose};
-
-    my $base = $argz{base};
-
-    # now locate the libraries.
-    #
-    # %found contains the abs_path results for each directory to 
-    # avoid double-including directories.
-    #
-    # note: loop short-curcuts for the (usually) list.
-
-    %found = ();
-
-    for( @{ $argz{ ignore } } )
-    {
-      if( my $dir = eval { abs_path catdir $_, $base } )
-      {
-        if( -d $dir )
-        {
-          $found{ $dir } = 1;
-        }
-      }
-    }
+    %argz
 };
 
 sub import
 {
-    &$handle_args;
+    # map used here for -T checks. everything from find_dirs has
+    # already been through abs_path, not much more to check here.
 
-    my @libz = $find_libs->();
+    my %argz    = &$handle_args;
+    my @libz    
+    = map
+    {
+        my ( $path ) = m{ (.+) }x; 
 
-    # HAK ALERT: the regex does nothing for security,
-    # just dodges -T. putting this down here instead
-    # of inside find_libs allows people to use saner
-    # untainting plans via find_libs.
+        ( $path =~ /\S/ )
+        ? $path
+        : ()
+    }
+    $find_dirs->( \%argz )
+    ;
 
-    @libz   = map { m{ (.+) }x } @libz;
+    my $verbose = $argz{ verbose };
+    my $caller  = caller;
 
-    my $caller = caller;
-
-    if( $verbose || defined $argz{print} )
+    if( $verbose || defined $argz{ print } )
     {
         local $\ = "\n";
         local $, = "\n\t";
 
         print STDERR "Found */$argz{ base }:", @libz
+        if $verbose;
     }
 
-    if( $argz{export} )
+    if( $argz{ export } )
     {
-        print STDERR join '', "\nExporting: @", $caller, '::', $argz{export}, "\n"
-        if $verbose;
-
-        # Symbol this is cleaner than "no strict" 
-        # for installing the array.
+        # this has to run in order to install variables that
+        # the caller is expecting to exist at runtime -- even
+        # if they are empty/undef at the end of it.
 
         my $ref = qualify_to_ref $argz{ export }, $caller;
 
-        *$ref = \@libz;
+        if( $verbose )
+        {
+            my $dest    = qualify $argz{ export }, $caller;
+
+            $argz{ scalar }
+            ? print STDERR "\nExporting: \$$dest\n"
+            : print STDERR "\nExporting: \@$dest\n"
+            ;
+        }
+
+        if( $argz{ scalar } )
+        {
+            *$ref 
+            = @libz
+            ? \$libz[0]
+            : \( my $a = '' )
+            ;
+        }
+        elsif
+        (
+            $argz{ append } 
+            and
+            my $ary = *{ $ref }{ ARRAY }
+        )
+        {
+            push @$ary, @libz;
+        }
+        else
+        {
+            *$ref = \@libz
+        }
     }
 
-    if( defined $argz{ p5lib } )
+    if( @libz )
     {
-        # stuff the lib's found at the front of $ENV{ PERL5LIB }
+        # no 'else', these are not exclusive
 
-        ( substr $ENV{ PERL5LIB }, 0, 0 ) = join ':', @libz, ''
-        if @libz;
+        if( defined $argz{ p5lib } )
+        {
+            # stuff the lib's found at the front of $ENV{ PERL5LIB }
+            # yes, virginia, substr is an lvalue -- and saner than
+            # dealing with \Q and a regex on arbitrary paths.
 
-        print STDERR "\nUpdated PERL5LIB:\t$ENV{ PERL5LIB }\n"
-        if $verbose;
-    }
+            ( substr $ENV{ PERL5LIB }, 0, 0 ) = join ':', @libz, '';
 
-    if( $argz{use} && @libz )
-    {
-        # this obviously won't work if lib ever depends 
-        # on the caller's package.
-        #
-        # it does avoids issues with -T blowing up on the
-        # old eval technique.
+            print STDERR "\nUpdated PERL5LIB:\t$ENV{ PERL5LIB }\n"
+            if $verbose;
+        }
 
-        require lib;
+        if( $argz{ use } )
+        {
+            # this obviously won't work if lib ever depends 
+            # on the caller's package.
 
-        lib->import( @libz );
+            require lib;
+
+            lib->import( @libz );
+        }
     }
 
     0
-};
+}
 
 # keep require happy
-
 1
-
 __END__
 
 =head1 NAME
 
-FindBin::libs_5_8 - stable version of code for older
-perl installations.
+FindBin::libs - locate and a 'use lib' or export 
+directories based on $FindBin::Bin.
 
 =head1 SYNOPSIS
 
-FindBin::libs locates and performs a 'use lib' for 
-directories along the path of $FindBin::Bin to automate 
-locating modules.  Uses File::Spec and Cwd's abs_path to 
-accomodate multiple O/S and redundant symlinks.
-
     # search up $FindBin::Bin looking for ./lib directories
     # and "use lib" them.
 
@@ -516,11 +541,24 @@ that
 
 will search for directories named "altlib" and "use lib" them.
 
-=head3 Exporting a variable: 'export'
+=head3 Exporting a variable: "export", "scalar", "append"
 
-The 'export' option will push an array of the directories found
-and takes an optional argument of the array name, which defaults 
-to the basename searched for:
+=over 4
+
+=item "export"
+
+This installs the results of locating directories into the caller's 
+space. Without any argument, export pushes out a variable named after 
+the located [sub]dir; an argument can be supplied to give the variable 
+name. Without the "scalar" option, the exported variable will be an 
+array in increasing order of "distance" (i.e., "up" the file tree); 
+with the "scalar" option only the first (i.e., "nearest") path is 
+exported.
+
+If "export" is given then "nouse" is assumed; using both leaves the 
+variable exported and its contents handed to "use lib".
+
+For example:
 
     use FindBin::libs qw( export );
 
@@ -550,7 +588,7 @@ exports @meta while
 
     use FindBin::libs qw( export=metadirs base=meta );
 
-exports @metadirs.
+exports @metadirs as a list of paths ending in "/meta".
 
 The use and export switches are not exclusive:
 
@@ -559,21 +597,85 @@ The use and export switches are not exclusive:
 will locate "lib" directories, use lib them, and export 
 @mylibs into the caller's package. 
 
+=item "scalar"
+
+Only searches for the first directory, which is exported (or
+overwritten) as a scalar rather than array. For example, if
+a project directory has ./bin and ./etc dir's then #! code in
+bin with
+
+    use FindBin::libs qw( export scalar base=etc );
+
+will have an $etc variable with the absolute path to ./bin/../etc.
+For configuration varibles this is usually what you want and allows
+for "$etc/Foo.conf" rather than "$etc[0]/Foo.conf".
+
+=item "append"
+
+Sometimes it's simpler to accumulate multiple searches into a 
+single array. Say for ./etc dir's in collection of standard
+locations.
+
+In that case:
+
+    use FindBin::libs qw( export=etc base=foo subdir=etc );
+    use FindBin::libs qw( export=etc base=bar subdir=etc append );
+
+produces something like
+
+    (
+        /path/to/foo/etc
+        /path/to/bar/etc
+    )
+
+without append @etc will have only ./bar/etc since the array would
+be overwritten with each call to import.
+
+=back
+
 =head3 Subdirectories
 
 The "subdir" and "subonly" settings will add or 
 exclusively use subdir's. This is useful if some
 of your lib's are in ../lib/perl5 along with 
-../lib (subdir=perl5) or all of the lib's are 
-in ../lib/perl5 (subonly=perl5).
+../lib or all of the lib's are in ../lib/perl5.
+
+These could be handled with:
+
+    use FindBin::libs;
+    use FindBin::libs qw( subdir=perl5 subonly );
+
+which uses the "lib" dir's along with any lib/perl5 dirs.
 
 This can also be handy for locating subdir's used
 for configuring packages:
 
     use FindBin::libs qw( export base=config subonly=mypackage );
 
-Will leave @config with any "mypackage" holding
-any "mypackage" subdir's.
+Will leave @config containing any mypackage dir's found up
+the tree, nearest to closest.
+
+The array format is convienent for locating configuration files
+shared between projects in separate, sibling directories. For
+example given:
+
+    ./proj/Foo/etc
+    ./proj/etc
+
+with
+
+    use FindBin::libs qw( export subdir=etc subonly )
+
+will export @etc with qw( ../proj/Foo/etc ../proj/etc ) in lexical
+order by distance from the #! code. At that point
+
+    use List::Util qw( first );
+
+    my $path = first { -e "$_/Global.config" } @etc;
+
+will locate the nearest "Global.confg" file. Note that this is 
+not the same as using "scalar" since that will export 
+$etc with only ./Foo/etc. 
 
 =head3 Setting PERL5LIB: p5lib
 
@@ -604,8 +706,6 @@ to something like:
 This can make controlling the paths used simpler and avoid
 the use of symlinks for some testing (see examples below).
 
-Note that "p5lib" and "nouse" are proably worth 
-
 =head2 Skipping directories
 
 By default, lib directories under / and /usr are
@@ -653,7 +753,7 @@ having to modify a single line of code.
 Say your sandbox is in ./sandbox and you are currently
 working in ./sandbox/projects/package/bin on a perl
 executable. You may have some number of modules that
-are specific -- or customized -- for this package, 
+are specific -- or customized -- for this pacakge, 
 share some modules within the project, and may want 
 to use company-wide modules that are managed out of 
 ./sandbox in development. All of this lives under a 
@@ -793,7 +893,7 @@ allow:
 
 to check the code.
 
-=back 
+=back
 
 =head1 Notes
 
@@ -810,7 +910,7 @@ may be limited to *NIX or at least less portable.
 
 PERL5LIB can be used to accomplish the same directory
 lookups as FindBin::libs.  The problem is PERL5LIB often
-contains absolte paths and does not automatically change
+contains absolute paths and does not automatically change
 depending on where tests are run. This can leave you 
 modifying a file, changing directory to see if it works
 with some other code and testing an unmodified version of 
@@ -852,7 +952,7 @@ in the executable path (or cwd) and uses that.
 
 The main problem here is that if the anchor ever changes
 (e.g., when moving code between projects or relocating 
-directories now that SVN supports it) the path often has
+directories) the path often has
 to change in multiple files. The regex also may have to
 support multiple platforms, or be broken into more complicated
 File::Spec code that probably looks pretty much like what
@@ -915,10 +1015,14 @@ where '-e' would have succeded.
 This is used for portability in dis- and re-assembling 
 directory paths based on $FindBin::Bin.
 
-=item libs_curr.pm
+=item Older code.
 
-This is installed if $^V indicates that the running
-perl is >= v5.10.
+There are versions of the code for Perl from releases
+v5.8, 5.14, and 5.28. Makefile.PL will pick the most recent
+version it can based on $^V at Configure time. As of this
+writing Perl is at 5.32, which would pick the 5.28 release. 
+
+The version-specific code is in ./version/<perl-v-string.
 
 =back
 
@@ -947,26 +1051,22 @@ bug, this is obviously a somewhat kludgy workaround
 and should be removed (with an added test for a 
 working version) once the File::Spec is fixed.
 
-=item 
+Note that I have no way to test this on VMS at this
+point so the only thing I can think of is declaring
+the fix a feature.
 
-The hack for prior-to-5.12 versions of perl is 
-messy, but is the only I've found that works for
-the moment on *NIX, VMS, and MSW. I am not sure
-whether any of these systems are normally configured
-to share perl modules between versions. If the 
-moduels are not shared on multiple platforms then
-I can make this work by managing the installation
-rather than checking this every time at startup.
-
-For the moment, at least, this seems to work.
-
-=back 
+=back
 
 =head1 SEE ALSO
 
-=head2 Module::FromPerlVer
+=over 4
 
-Explains where the installed version comes from.
+=head2 Cwd
+
+A bug in Cwd was fixed with 3.73. That should, hopefully,
+fix the issue with FB::l croaking perl during testing.
+If it does not then I will have to modify the sanity check
+for using abs_path vs. rel2abs.
 
 =head1 AUTHOR
 
@@ -974,11 +1074,11 @@ Steven Lembark, Workhorse Computing <lembark@wrkhors.com>
 
 =head1 COPYRIGHT
 
-Copyright (C) 2003-2012, Steven Lembark, Workhorse Computing.
-This code is released under the same terms as Perl-5.24
+Copyright (C) 2003-2022, Steven Lembark, Workhorse Computing.
+This code is released under the same terms as Perl-5.28
 or any later version of Perl.
 
 =head1 LICENSE
 
-This code is released under the same terms as Perl-5.24
+This code is released under the same terms as Perl-5.34
 or any later version of Perl.
